#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <gtk/gtk.h>

char *colors[] = {"Black", "Blue", "Red", "Yellow"};

typedef enum {BLACK, BLUE, RED, YELLOW} color_t;

typedef struct card {
    color_t color;
    int number;
} card_t;

#define DECK_MAX_CARDS 104

typedef struct deck {
    card_t cards[DECK_MAX_CARDS];
    int num_cards;
} deck_t;

#define CARD_GROUP_MAX_CARDS DECK_MAX_CARDS

typedef struct card_group {
    card_t cards[CARD_GROUP_MAX_CARDS];
    int num_cards;
} card_group_t;

#define BOARD_MAX_CARD_GROUPS (DECK_MAX_CARDS / 3)

typedef struct board {
    card_group_t groups[BOARD_MAX_CARD_GROUPS];
    int num_groups;
} board_t;

typedef struct player {
    card_group_t hand;
} player_t;

#define GAME_MAX_PLAYERS 4
#define GAME_WINDOW_DEFAULT_WIDTH  800
#define GAME_WINDOW_DEFAULT_HEIGHT 600

typedef struct game {
    player_t players[GAME_MAX_PLAYERS];
    int num_players;
    board_t board;
    deck_t deck;
} game_t;

static void card_print(card_t card)
{
    printf("%6s %2d\n", colors[card.color], card.number + 1);
}

static void card_group_init(card_group_t *card_group)
{
    card_group->num_cards = 0;
}

static void board_init(board_t *board)
{
    int i;
    board->num_groups = 0;
    
    for (i = 0; i <= BOARD_MAX_CARD_GROUPS; ++i) 
    {
	card_group_init(&board->groups[i]);
    }
}

static void player_init(player_t *player)
{
    card_group_init(&player->hand);
}


static int card_compare(card_t *card_one, card_t *card_two)
{
    if (card_one->number < card_two->number)
	return -1;
    if (card_one->number == card_two->number)
	return 0;
    /* At this point, we know that this must be true:
     * if (card_one->number > card_two->number)
     */
    return 1;
}

static int card_group_is_run(card_group_t *card_group)
{
    int i;
    int lowest = 14, highest = 0;
    if (card_group->num_cards > 13 || card_group->num_cards < 3)
    {
	return 0;
    }
    for (i = 0; i < card_group->num_cards - 1; ++i)
    {
	if (card_group->cards[i].color != card_group->cards[i + 1].color)
	{
	    return 0;
	}
	if (card_group->cards[i].number > highest)
	{
	    highest = card_group->cards[i].number;
	}
	if (card_group->cards[i].number < lowest)
	{
	    lowest = card_group->cards[i].number;
	}
    }
    if (highest - lowest != card_group->num_cards - 1)
    {
	return 0;
    }
    return 1;
}

static int card_group_is_set(card_group_t *card_group)
{
    int i;
    color_t seen_color[card_group->num_cards];
    
    if (card_group->num_cards > 4 || card_group->num_cards < 3)
    {
	return 0;
    }
    for (i = 0; i < card_group->num_cards - 1; ++i) 
    {
	if (card_group->cards[i].number != card_group->cards[i + 1].number)
	{
	    return 0;
	}
    }
    seen_color[i] = card_group->cards[i].color;
    for (i = 0; i < card_group->num_cards; ++i)
    {
	seen_color[card_group->cards[i].color]++;
	if (seen_color[card_group->cards[i].color] > 1)
	{
	    return 0;
	}
    }
    return 1;
}

static void deck_deal(game_t *game, deck_t *deck)
{
    card_t temp;
    int rand_card;
    int i, j, newline;
    
    printf ("How many players(1-4) should I deal in? ");
    game->num_players = getchar();
    if (game->num_players == EOF)
    {
	printf ("\nGoodbye.\n");
	exit (1);
    }
    newline = getchar();   
    game->num_players -= '0';
    
    for (i = 0; i < game->num_players; ++i)
    {
	for (j = 0; j < 14; ++j)
	{
	    rand_card = ((deck->num_cards + 1.0) * rand()) / (RAND_MAX + 1.0);
	    temp = deck->cards[rand_card];
	    deck->cards[rand_card] = deck->cards[deck->num_cards - 1];
	    game->players[i].hand.cards[j] = temp;
	    deck->num_cards -= 1;
	    game->players[i].hand.num_cards += 1;
	}
    }
    printf ("Game dealt for %d player(s)\n", game->num_players);
}

static void deck_init(deck_t *deck)
{
    int h, i, j;  
    deck->num_cards = 0;
    for (h = 0; h <= 1; ++h)
    {
	for (i = 0; i <= 3; ++i) 
	{
	    for (j = 0; j <= 12; ++j) 
	    {
		deck->cards[j + (i * 13) + (h * 52)].color = i;
		deck->cards[j + (i * 13) + (h * 52)].number = j;
		deck->num_cards += 1;
		printf ("There are %d tiles in the deck\n", deck->num_cards);
	    }
	}
    }
} 

static void deck_shuffle(deck_t *deck)
{
    card_t temp;
    int rand_card;
    int last;
    for (last = deck->num_cards; last > 0; --last)
    {
	rand_card = ((last + 1.0) * rand()) / (RAND_MAX + 1.0);
	temp = deck->cards[rand_card];
	deck->cards[rand_card] = deck->cards[last - 1];
	deck->cards[last - 1] = temp;
    }
}

static void deck_print(deck_t *deck)
{
    int h, i, j;  
    for (h = 0; h <= 1; ++h)
    {
	for (i = 0; i <= 3; ++i)
	{
	    for (j = 0; j <= 12; ++j)
	    {
		card_print(deck->cards[j + (i * 13) + (h * 52)]);
	    }
	}
    }
    printf ("There are %d tiles in the deck\n" , deck->num_cards);
}

static void hand_print(game_t *game)
{
    int i;
    for (i = 0; i < game->players[0].hand.num_cards; ++i)
    {
	card_print(game->players[0].hand.cards[i]);
    }
}

static void game_init(game_t *game)
{
    int i;
    game->num_players = 0;
    
    for (i = 0; i < GAME_MAX_PLAYERS; ++i)
    {
	player_init(&game->players[i]);
	game->num_players += 1;
    }
    
    board_init(&game->board);
    deck_init(&game->deck);
    deck_shuffle(&game->deck);
}

static gboolean on_expose_event (GtkWidget *widget, GdkEventExpose *event, game_t *game)
{
    printf ("I should be drawing something here\n");

    return TRUE;
}

static gboolean on_key_press_event (GtkWidget *widget, GdkEventKey *event, game_t *game)
{
    printf ("You pressed key %d\n", event->keyval);

    return TRUE;
}

static gboolean on_button_press_event (GtkWidget *widget, GdkEventButton *event, game_t *game)
{
    printf ("You pressed button %d\n", event->button);

    return TRUE;
}

static gboolean on_button_release_event (GtkWidget *widget, GdkEventButton *event, game_t *game)
{
    printf ("You released button %d\n", event->button);

    return TRUE;
}

int main(int argc, char *argv[])
{
    GtkWidget *window;
    game_t game;
    
    srand(time(NULL));
    
    game_init(&game);
    deck_print(&game.deck);
    deck_deal(&game, &game.deck);
    hand_print(&game);
    deck_print(&game.deck);

    gtk_init (&argc, &argv);

    /* Create a new window */
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_default_size (GTK_WINDOW (window),
				 GAME_WINDOW_DEFAULT_WIDTH,
				 GAME_WINDOW_DEFAULT_HEIGHT);

    gtk_widget_set_events (window,
			   GDK_EXPOSURE_MASK |
			   GDK_KEY_PRESS_MASK | 
			   GDK_BUTTON_PRESS_MASK | 
			   GDK_BUTTON_RELEASE_MASK);

    g_signal_connect (G_OBJECT (window), "delete_event",
		      G_CALLBACK (gtk_main_quit), NULL);
    g_signal_connect (G_OBJECT (window), "expose_event",
		      G_CALLBACK (on_expose_event), &game);
    g_signal_connect (G_OBJECT (window), "key_press_event",
		      G_CALLBACK (on_key_press_event), &game);
    g_signal_connect (G_OBJECT (window), "button_press_event",
		      G_CALLBACK (on_button_press_event), &game);
    g_signal_connect (G_OBJECT (window), "button_release_event",
		      G_CALLBACK (on_button_release_event), &game);

    gtk_widget_show_all (window);
    gtk_main ();

    return 0;

}
